import apt_pkg
import argparse
import collections
import copy
import glob
import os
import sys

def find_release(pkg: apt_pkg.Version) -> apt_pkg.Version | None:
    """Finds the version of the package in the release pocket. Sort of obsolete now since we use a filtered sources.list."""
    for version in pkg.version_list:
        for file, index in version.file_list:
            if "-" not in file.archive:
                return version

def release_universe(cache: apt_pkg.Cache) -> list[apt_pkg.Version]:
    """Finds all versions in the release pocket. Sort of obsolete, since we use a filtered sources.list"""
    versions_in_release = []
    for pkg in cache.packages:
        v = find_release(pkg)
        if v:
            versions_in_release.append(v)
    return versions_in_release

def get_dependencies(ver: apt_pkg.Version) -> list[apt_pkg.Version]:
    """Get all dependencies of the version object."""
    res = []
    for type, or_groups in ver.depends_list.items():
        for or_group in or_groups:
            for dep in or_group:
                if dep.dep_type_enum not in (apt_pkg.Dependency.TYPE_DEPENDS, apt_pkg.Dependency.TYPE_PREDEPENDS):
                    continue
                for target in dep.all_targets():
                    res.append(target)
    return res

def get_build_dependencies(cache: apt_pkg.Cache, srcrecords: apt_pkg.SourceRecords) -> list[apt_pkg.Version]:
    """Get all build dependencies for the current source package that srcrecords is located at."""
    res = []
    for type, or_groups in srcrecords.build_depends.items():
        if type.startswith("Build-Conflicts"):
            continue
        assert type in ["Build-Depends", "Build-Depends-Indep", "Build-Depends-Arch"]
        for or_group in or_groups:
            for name, op, ver in or_group:
                try:
                    pkg = cache[name]
                except KeyError:
                    continue
                for version in pkg.version_list:
                    res.append(version)
                for _name, _verstr, version in pkg.provides_list:
                    res.append(version)
    return res

def remove_leaves(depends: dict[str, set[str]], reverse_depends: dict[str, set[str]], in_proposed, args, i):
    """Iterate over the packages to find any that do not have reverse dependencies and drop them."""
    leaves = []
    # `depends` is the full package universe due to how the code works. It only contains binaries,
    # so we don't end up removing `src:` packages (which are used to hold build-depends).
    for package in sorted(depends):
        rdeps = reverse_depends.get(package, [])
        may_remove = not args.proposed or package in in_proposed
        if not rdeps and may_remove:
            leaves.append(package)
        elif args and args.log and i == 0:
            print(f"keep {package} rdeps={sorted(rdeps)} may_remove={may_remove}", file=args.log)

    for package in leaves:
        # Remove the leaf package
        for dep in depends[package]:
            reverse_depends[dep].remove(package)
        del depends[package]
    return leaves

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--proposed", help="Only packages with a newer source in proposed may be removed", action="store_true")
    parser.add_argument("--proposed-missing", help="Only packages missing in proposed may be removed", action="store_true")
    parser.add_argument("--log", type=argparse.FileType("w"))
    args = parser.parse_args()

    apt_pkg.config["dir"] = os.path.abspath("root")
    apt_pkg.config["dir::etc::sourcelist"] = "release-only.list"
    apt_pkg.config["apt::architecture"] = "armhf"
    apt_pkg.init()

    cache = apt_pkg.Cache(progress=None)
    srcrecords = apt_pkg.SourceRecords()
    pkgrecords = apt_pkg.PackageRecords(cache)

    universe = release_universe(cache)
    reverse_depends : dict[str, set[str]] = collections.defaultdict(set)
    depends : dict[str, set[str]] = collections.defaultdict(set)

    # Calculate all packages missing a build in proposed (listed in source package - those that are built in there)
    pending_proposed : set[str] = set()
    for proposed_sources in glob.glob("root/var/lib/apt/lists/*proposed*Sources"):
        for source in apt_pkg.TagFile(proposed_sources):
            for package in (p.strip() for p in source["Binary"].split(",")):
                pending_proposed.add(package)

    if args.proposed_missing:
        for proposed_binaries in glob.glob("root/var/lib/apt/lists/*proposed*Packages"):
            for binary in apt_pkg.TagFile(proposed_binaries):
                try:
                    pending_proposed.remove(binary["Package"])
                except KeyError:
                    pass    # Why can we have a binary in proposed without a source package

    # Source packages are reverse-dependencies only, we don't add them to the depends list
    while srcrecords.step():
        deps = get_build_dependencies(cache, srcrecords)
        name = "src:" + srcrecords.package
        for dep in deps:
            reverse_depends[dep.parent_pkg.name].add(name)

    # Packages listed in a task are "seeded", they must never be removed.
    for binaries in glob.glob("root/var/lib/apt/lists/*Packages"):
            for binary in apt_pkg.TagFile(binaries):
                name = binary["Package"]
                tasks = binary.get("Task")
                if tasks is None:
                    continue

                for task in (t.strip() for t in tasks.split(",")):
                    if task == name or "ubuntu-" + task == name:
                        reverse_depends[name].add("task:" + task)

    # Expand the dependency tree in the universe
    for ver in universe:
        deps = get_dependencies(ver)
        name = ver.parent_pkg.name
        depends[name] = set()
        for dep in deps:
            depends[name].add(dep.parent_pkg.name)
            reverse_depends[dep.parent_pkg.name].add(name)

    # Print the report
    print("# report: ", " ".join(sys.argv))
    if args.log:
        print("# log:", args.log.name)
    print("# architecture-specific leaf (i=0) and to-be-leaf (i > 0) binary packages")
    print("#")
    print("# Packages starting with 0 can be removed immediately, packages with i > 0 become removable if all in i-1...0 have been removed ")
    print("# We check both Depends and Build-Depends-* (and PreDepends), and seeded packages are never removable (by presence of Task: field)")
    print("# Provides are respected, of course")
    if args.proposed_missing:
        print("#")
        print("# This shows only leave packages for which a newer source package in proposed is missing a build")
    elif args.proposed:
        print("#")
        print("# This shows only leave packages for which a newer source package exists in proposed")
    print()
    leaves = []
    previous_leaves = set()
    original_reverse_depends = copy.deepcopy(reverse_depends)
    max_package_name_len = max(map(len, depends))
    for i in range(len(depends)):
        leaf = None
        new_leaves = set()
        for leaf in remove_leaves(depends, reverse_depends, pending_proposed, args, i):
            renderleaf = leaf + ":all" if cache[leaf].version_list[0].arch == "all" else leaf
            if i != 0:
                print(("%d %-" + str(max_package_name_len) + "s # reverse-depends removed in i=%d: %s") % (i, renderleaf, i - 1, " ".join(original_reverse_depends[leaf] & previous_leaves)))
            else:
                print(i, renderleaf)
            leaves.append(leaf)
            new_leaves.add(leaf)
        previous_leaves = new_leaves
        if leaf is None:
            break
    print()

main()
